
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.4. Last modification 28/05/08
*)
(**
transcript.ml
@author Yann GLOUCHE
@version 1.0
@since 12/12/2005
*)
open Utils_simulator;;
open Variables;;
	
(** Transcription IF to HLPSL, and HLPSL to IF *)

(** contains all the transcription of a message from a HLPSL format send by an  agent to Intermediate Format, and from Intermediate Format to HLPSL format *)
let transcript_table_sender = (Hashtbl.create 600 : (string, string) Hashtbl.t);;
(** 
transcript a message "mess" in HLPSL send by an agent to Intermediate Format, and transcript a message "mess" in Intermediate Format to HLPSL.
@param mess is a message in HLPSL or Intermediate Format
@return string, then message is in HLPSL or Intermediate Format
*)
let transcript_sender (mess : string) : string =
	let add mess1 mess2 =
		if(not(Hashtbl.mem transcript_table_sender mess2) && (mess=mess1 || not(Hashtbl.mem transcript_table_sender mess1)))
		then(
		Hashtbl.add transcript_table_sender mess1 mess2;
		if(mess1<>mess2) then Hashtbl.add transcript_table_sender mess2 mess1;
		);
		mess2
	in
	let rec aux mess =
		if(Hashtbl.mem transcript_table_sender mess) then Hashtbl.find transcript_table_sender mess
		else(
			let l_mess = decompose mess in
			(add mess
				(if(l_mess=[]) then mess
				else if((car l_mess)="pair") then (list2string "." (List.map (fun e -> aux e) (cdr l_mess)))
				else if((car l_mess)="scrypt" || (car l_mess)="crypt") then (
						let tmp=decompose (cadr l_mess) in
						if(tmp<>[] && (car tmp)="pair")
						then "{"^(aux(caddr l_mess))^"}_{"^(aux (cadr l_mess))^"}"
						else "{"^(aux(caddr l_mess))^"}_"^(aux (cadr l_mess));
					)
				else if((car l_mess)="apply") then (aux(cadr l_mess))^"("^ (list2string "," (List.map aux (cddr l_mess)) ) ^")"
				else if((car l_mess)="exp") then (car l_mess)^"("^ (list2string "," (List.map aux (cdr l_mess)) ) ^")"
				else if((car l_mess)="xor") then (car l_mess)^"("^ (list2string "," (List.map aux (cdr l_mess)) ) ^")"
				else if((car l_mess)="inv") then (car l_mess)^"("^ (aux (cadr l_mess)) ^")"
				else 	let tmp = 	(list2string ","	
								(List.map (fun s -> 	add s (aux s)) 
								(cdr l_mess) ) 
							) in
					(add (car l_mess) (aux (car l_mess)))^"("^tmp^")";
			))	
		);
	in
	aux mess
;;

(** contains all the transcription of a message receives by an agent from a HLPSL format to Intermediate Format, and from Intermediate Format to HLPSL format *)
let transcript_table_receiver = (Hashtbl.create 600 : (string, string) Hashtbl.t);;
(** 
transcript a message "mess" receives by an agent in HLPSL to Intermediate Format, and transcript a message "mess" in Intermediate Format to HLPSL.
@param mess is a message in HLPSL or Intermediate Format
@return string, then message is in HLPSL or Intermediate Format
*)
let transcript_receiver (mess : string) : string =
	let add mess1 mess2 =
		if(not(Hashtbl.mem transcript_table_receiver mess2) && (mess=mess1 || not(Hashtbl.mem transcript_table_receiver mess1)))
		then(
		Hashtbl.add transcript_table_receiver mess1 mess2;
		if(mess1<>mess2) then Hashtbl.add transcript_table_receiver mess2 mess1;
		);
		mess2
	in
	let rec aux mess =
		if(Hashtbl.mem transcript_table_receiver mess) then Hashtbl.find transcript_table_receiver mess
		else(
			let l_mess = decompose mess in
			(add mess
				(if(l_mess=[]) then(
					if(is_a_test_var mess) then get_name_of_test_var mess
					else if(is_a_var mess) then mess^"'"
					else mess
				)
				else if((car l_mess)="pair") then (list2string "." (List.map (fun e -> aux e) (cdr l_mess)))
				else if((car l_mess)="scrypt" || (car l_mess)="crypt") then (
						let tmp=decompose (cadr l_mess) in
						if(tmp<>[] && (car tmp)="pair")
						then "{"^(aux(caddr l_mess))^"}_{"^(aux (cadr l_mess))^"}"
						else "{"^(aux(caddr l_mess))^"}_"^(aux (cadr l_mess));
					)
				else if((car l_mess)="apply") then (aux(cadr l_mess))^"("^ (list2string "," (List.map aux (cddr l_mess)) ) ^")"
				else if((car l_mess)="exp") then (car l_mess)^"("^ (list2string "," (List.map aux (cdr l_mess)) ) ^")"
				else if((car l_mess)="xor") then (car l_mess)^"("^ (list2string "," (List.map aux (cdr l_mess)) ) ^")"
				else if((car l_mess)="inv") then (car l_mess)^"("^ (aux (cadr l_mess)) ^")"
				else 	let tmp = 	(list2string ","	
								(List.map (fun s -> 	add s (aux s)) 
								(cdr l_mess) ) 
							) in
					(add (car l_mess) (aux (car l_mess)))^"("^tmp^")";
			))	
		);
	in
	aux mess
;;
(**
return a message with all the value of variables contained in this message
@param name is the sender of the message "mess"
@param mess is a message send by the agent "name"
@return the message "mess" with all the value of the variables contained in the message
*)
let transcript_to_real name (mess : string) : string =
	let rec aux mess =
		let l_mess = decompose mess in
		(if(l_mess=[]) then(
			if(is_a_var mess) then aux (get_value_of_var name mess)
			else mess
		)
		else if((car l_mess)="pair") then (list2string "." (List.map (fun e -> aux e) (cdr l_mess)))
		else if((car l_mess)="scrypt" || (car l_mess)="crypt") then (
				let tmp=decompose (cadr l_mess) in
				if(tmp<>[] && (car tmp)="pair")
				then "{"^(aux(caddr l_mess))^"}_{"^(aux (cadr l_mess))^"}"
				else "{"^(aux(caddr l_mess))^"}_"^(aux (cadr l_mess));
			)
		else if((car l_mess)="apply") then (aux(cadr l_mess))^"("^ (list2string "," (List.map aux (cddr l_mess)) ) ^")"
		else if((car l_mess)="exp") then (car l_mess)^"("^ (list2string "," (List.map aux (cdr l_mess)) ) ^")"
		else if((car l_mess)="xor") then (car l_mess)^"("^ (list2string "," (List.map aux (cdr l_mess)) ) ^")"
		else if((car l_mess)="inv") then (car l_mess)^"("^ (aux (cadr l_mess)) ^")"
		else 	let tmp = 	(list2string "," (cdr l_mess) 
					) in
			(aux (car l_mess))^"("^tmp^")";
		)
	in
	aux mess
;;


(** contains all the transcription of a message from a HLPSL format send by an  agent to Intermediate Format, and from Intermediate Format to HLPSL format  with all the value of variables contained in this message *)
let transcript_table_for_intruder = (Hashtbl.create 600 : ((string * string), string) Hashtbl.t);;
(** 
transcript a message "mess" receives and show the tests values in this messages by an agent in HLPSL to Intermediate Format, and transcript a message "mess" in Intermediate Format to HLPSL.
@param name is the sender of the message "mess"
@param mess is a message send by the agent "name"
@param mess is a message in HLPSL or Intermediate Format
@return string, then message is in HLPSL or Intermediate Format
*)
let transcript_receiver_for_intruder (name : string) (mess : string) : string =
	let add mess1 mess2 =
		if(not(Hashtbl.mem transcript_table_for_intruder (name,mess2)) && (mess=mess1 || not(Hashtbl.mem transcript_table_for_intruder (name,mess1))))
		then(
		Hashtbl.add transcript_table_for_intruder (name,mess1) mess2;
		if(mess1<>mess2) then Hashtbl.add transcript_table_for_intruder (name,mess2) mess1;
		);
		(mess1,mess2)
	in
	let rec aux mess =
		if(Hashtbl.mem transcript_table_for_intruder (name,mess)) 
		then (

		  let res1 = Hashtbl.find transcript_table_for_intruder (name,mess) in
		  let res2 = Hashtbl.find transcript_table_for_intruder (name,res1) in
		  (res2,res1)
		)
		else(
			let l_mess = decompose mess in
				if(l_mess=[]) then (
				if(is_a_test_var mess) then (
					let value = get_value_of_var name mess in
					add value value
					)
				else if(is_a_var mess) then  add mess ((get_name_of_var mess)^"'")
				else add mess mess
				)
				else if((car l_mess)="pair") then( 
				  let tm = List.map aux (cdr l_mess) in
				  add ((car l_mess)^"("^(list2string "," (List.map fst tm))^")") (list2string "." (List.map snd tm))
				  (*let (m1, t1) = aux (cadr l_mess) in
				  let (m2, t2) = aux (caddr l_mess) in
				    add ((car l_mess)^"("^m1^","^m2^")")  (t1^"."^t2) *)
				)
				else if((car l_mess)="scrypt" || (car l_mess)="crypt") then (
						let tmp=decompose (cadr l_mess) in
						if(tmp<>[] && (car tmp)="pair")
						then (
						  let (m1, t1) = aux (cadr l_mess) in
						  let (m2, t2) = aux (caddr l_mess) in
						  add ((car l_mess)^"("^m1^","^m2^")") ("{"^t2^"}_{"^t1^"}")
						)
						else (
						  let (m1, t1) = aux (cadr l_mess) in
						  let (m2, t2) = aux (caddr l_mess) in
						  add ((car l_mess)^"("^m1^","^m2^")") ("{"^t2^"}_"^t1)
					)
				)
				else if((car l_mess)="apply") then (
				  let tm = List.map aux (cdr l_mess) in
				  add ((fst (car tm))^"("^(list2string "," (List.map fst (cdr tm)))^")") ((snd (car tm))^"("^(list2string "," (List.map snd (cdr tm)))^")")
				)
				else if((car l_mess)="exp") then (
				  let tm = List.map aux (cdr l_mess) in
				  add ((car l_mess)^"("^(list2string "," (List.map fst tm))^")") ((car l_mess)^"("^(list2string "," (List.map snd tm))^")")
				)
				else if((car l_mess)="xor") then (
				  let tm = List.map aux (cdr l_mess) in
				  add ((car l_mess)^"("^(list2string "," (List.map fst tm))^")") ((car l_mess)^"("^(list2string "," (List.map snd tm))^")")
				)
				else if((car l_mess)="inv") then ( 
				  let (m1, t1) = aux (cadr l_mess) in
				    add ((car l_mess)^"("^m1^")")  ((car l_mess)^"("^t1^")") 
				)
				else 	(
				  let tm = List.map aux l_mess in
				  add ((fst (car tm))^"("^(list2string "," (List.map fst (cdr tm)))^")") ((snd (car tm))^"("^(list2string "," (List.map snd (cdr tm)))^")")
				)
		);
	in
	(snd (aux mess))
;;

(**
Compute the maximal length of the strings which represents the messages in the automata
@return the maxiamal length of the string whcich represents the messages (to send and to receive) in the automata
*)
let max_length_message automate = 
	max_length_string
	(List.flatten 
		(List.map 
			(fun (name,id,_,actionsL) ->
				List.flatten
				(List.map 
				(fun (initial,_,final) ->
				[(transcript_to_real (get_name_in_session name id) initial);(transcript_to_real (get_name_in_session name id) final)])
				actionsL)
			)
			automate))
;;